#!/usr/bin/perl
use lib "./UPnP/lib";
use UPnP::ControlPoint;
use Socket;
use IO::Select;
use IO::Handle;
use Data::Dumper;
use HTML::Parser;
use HTML::Entities;
use XML::Simple;
use HTTP::Daemon;
use HTML::Template;
use POSIX qw(strftime);
use strict;

###############################################################################
#CONFIG
$main::MAX_LOG_LEVEL  = 3;    # Lower, the less output
$main::HTTP_PORT      = 8001; # Port our fake http server listens on
$main::MAX_SEARCH     = 200;  # Max search results to return
$main::RENEW_SUB_TIME = 3600; # How often do we do a UPnP renew in seconds
$main::VERSION = "0.02";

$| = 1;


use SOAP::Lite maptype => {}; 

$SIG{INT} = "main::quit";
$SIG{PIPE} = sub {};
$SIG{CHLD} = sub {};

$Data::Dumper::Indent = 1;
@main::CALLBACKS = ();
@main::WAITINGAV = ();

main();

###############################################################################
# main
sub main {
    $main::daemon = HTTP::Daemon->new(LocalPort => $main::HTTP_PORT) || die;
    $main::cp = UPnP::ControlPoint->new ();
    my $search = $main::cp->searchByType("urn:schemas-upnp-org:device:ZonePlayer:1", \&main::search_cb);

    my @selsockets = $main::cp->sockets();
    @selsockets = (@selsockets, $main::daemon);
    my $select = IO::Select->new(@selsockets);


    if ( -f "musicdb.pl") {
        Log (0, "Loading Music DB");
        require "musicdb.pl";
        Log (0, "Finished Loading Music DB");
    } else {
        Log (0, "No Music DB - Click ReIndex after selecting a Zone");
    }
    renew_subscriptions();

    # MAIN LOOP
    while (1) {

        # Check the callbacks we have waiting
        my $timeout = 5;
        my $now = time;
        while ($#main::CALLBACKS >= 0) {
            if ($main::CALLBACKS[0][0] <= $now) {
                my($time, $callback) = @{shift @main::CALLBACKS};
                Log (0, "$callback");
                &$callback;
            } else {
                $timeout = $main::CALLBACKS[0][0] - $now;
                last;
            }
        }

        # Find if any sockets are ready for reading
        my @sockets = $select->can_read($timeout);

        # Call the handlers for the sockets
        for my $sock (@sockets) {
            if ($sock == $main::daemon) {
                my $c = $main::daemon->accept;
                my $r = $c->get_request;
                handle_http_request($c, $r);
            } else {
                $main::cp->handleOnce($sock);
            }
        }
    }
}
###############################################################################
sub add_callback {
    my ($time, $cb) = @_;

    @main::CALLBACKS = sort { $a->[0] <=> $b->[0] } @main::CALLBACKS, [$time, $cb];
}
###############################################################################
sub Log {
    my $level = shift;
    my $now = strftime "%e/%H%M%S", localtime;
    my @Level = ("crit", "notice", "info", "debug");

    return if ($level > $main::MAX_LOG_LEVEL);

    my ($package, $filename, $line) = caller(0);
    my ($x, $y, $z, $subroutine)    = caller(1);
    $subroutine = "*unknown*" if (!defined $subroutine);
    print "$now $Level[$level]: $filename:$line $subroutine @_\n";
}

###############################################################################
sub handle_http_request {
    my ($c, $r) = @_;

    my $uri = $r->uri;

    my $path = $uri->path;
    $path = "/index.html" if ($path eq "/");
    $path = "/index.html" if ($path =~ /\.\./);

    my %qf = $uri->query_form;
    Log (1, "URI: $uri");

    # File doesn't exist
    if ( ! -f "html/$path") {
        $c->send_error(HTTP::Status::RC_FORBIDDEN);
        $c->force_last_request;
        return;
    }

    # File isn't HTML
    if ( ! $path =~ /\.html/) {
        $c->send_file_response("html/$path");
        $c->force_last_request;
        return;
    }

    # 0 - not handled, 1 - handled send reply, 2 - handled routine will send reply
    my $response = 0;

    if ($qf{action}) {
        if ($qf{zone}) {
            $response = handle_http_zone_action($c, $r, $path);
        }
        if (! $response) {
            $response = handle_http_action($c, $r, $path);
        }
    }

    if ($response != 2) {
        send_http_response($c, $r, $path);
    }
}

###############################################################################
sub handle_http_zone_action {
my ($c, $r, $path) = @_;

    my %qf = $r->uri->query_form;

    if ($qf{action} eq "Remove") {
        avtransport_remove_track($main::ZONES{$qf{zone}}->{Location}, $qf{queue});
        push (@main::WAITINGAV, [$c, $r, $path]);
        return 2;
    } elsif ($qf{action} eq "RemoveAll") {
        avtransport_action($main::ZONES{$qf{zone}}->{Location}, "RemoveAllTracksFromQueue");
        push (@main::WAITINGAV, [$c, $r, $path]);
        return 2;
    } elsif ($qf{action} eq "Play") {
        avtransport_play($main::ZONES{$qf{zone}}->{Location});
    } elsif ($qf{action} =~ /(Next|Previous|Stop|Pause)/) {
        avtransport_action($main::ZONES{$qf{zone}}->{Location}, $qf{action});
        push (@main::WAITINGAV, [$c, $r, $path]);
        return 2;
    } elsif ($qf{action} eq "ShuffleOn") {
        avtransport_shuffle($main::ZONES{$qf{zone}}->{Location}, 1);
    } elsif ($qf{action} eq "ShuffleOff") {
        avtransport_shuffle($main::ZONES{$qf{zone}}->{Location}, 0);
    } elsif ($qf{action} eq "RepeatOn") {
        avtransport_repeat($main::ZONES{$qf{zone}}->{Location}, 1);
    } elsif ($qf{action} eq "RepeatOff") {
        avtransport_repeat($main::ZONES{$qf{zone}}->{Location}, 0);
    } elsif ($qf{action} eq "Seek") {
        avtransport_seek($main::ZONES{$qf{zone}}->{Location}, $qf{queue});
        push (@main::WAITINGAV, [$c, $r, $path]);
        return 2;
    } elsif ($qf{action} eq "MuteOn") {
        render_mute($main::ZONES{$qf{zone}}->{Location}, 1);
    } elsif ($qf{action} eq "MuteOff") {
        render_mute($main::ZONES{$qf{zone}}->{Location}, 0);
    } elsif ($qf{action} eq "MuchSofter") {
        render_volume_change($main::ZONES{$qf{zone}}->{Location}, -5);
    } elsif ($qf{action} eq "Softer") {
        render_volume_change($main::ZONES{$qf{zone}}->{Location}, -1);
    } elsif ($qf{action} eq "Louder") {
        render_volume_change($main::ZONES{$qf{zone}}->{Location}, +1);
    } elsif ($qf{action} eq "MuchLouder") {
        render_volume_change($main::ZONES{$qf{zone}}->{Location}, +5);
    } elsif ($qf{action} eq "AddMusic") {
        avtransport_add($main::ZONES{$qf{zone}}->{Location}, $qf{mpath});
        push (@main::WAITINGAV, [$c, $r, $path]);
        return 2;
    } elsif ($qf{action} eq "PlayMusic") {
        avtransport_action($main::ZONES{$qf{zone}}->{Location}, "RemoveAllTracksFromQueue");
        avtransport_add($main::ZONES{$qf{zone}}->{Location}, $qf{mpath});
        avtransport_play($main::ZONES{$qf{zone}}->{Location});
        push (@main::WAITINGAV, [$c, $r, $path]);
        return 2;
    } else {
        return 0;
    }
    return 1;
}

###############################################################################
sub handle_http_action {
my ($c, $r, $path) = @_;

    my %qf = $r->uri->query_form;

    if ($qf{action} eq "ReIndex") {
        reindex();
    } else {
        return 0;
    }
    return 1;
}

###############################################################################
sub send_http_response {
my ($c, $r, $path) = @_;

    my %qf = $r->uri->query_form;

    # One of ours templates, now fill in the parts we know
    my $template = HTML::Template->new(filename => "html/$path",
                                       die_on_bad_params => 0,
                                       cache => 1);

    $template->param("VERSION" => $main::VERSION);

    # There is a ZONES_LOOP so gather all the information about the zones
    if ($template->query(name => "ZONES_LOOP")) {
        my @loop_data = ();
        foreach my $zone (keys %main::ZONES) {
            my %row_data;
            $row_data{ZONE_NAME} = $main::ZONES{$zone}->{ZoneName};
            $row_data{ZONE_ARG}  = "zone=$zone";
            push(@loop_data, \%row_data);
        }
        $template->param("ZONES_LOOP" => \@loop_data);
    }
    
    # There was a zone param, incase the template wants the zone name
    if ($qf{zone}) {
        $template->param("ACTIVE_ZONE" => $main::ZONES{$qf{zone}}->{ZoneName},
                         "ACTIVE_ZONEID" => $qf{zone});
        my $curtrack = $main::AV{"$qf{zone},CurrentTrackMetaData"};

        if ($curtrack) {
            $template->param("ACTIVE_NAME"      => $curtrack->{item}->{"dc:title"},
                             "ACTIVE_ARTIST"    => $curtrack->{item}->{"dc:creator"},
                             "ACTIVE_ALBUM"     => $curtrack->{item}->{"upnp:album"},
                             "ACTIVE_TRACK_NUM" => $main::AV{"$qf{zone},CurrentTrack"},
                             "ACTIVE_TRACK_TOT" => $main::AV{"$qf{zone},NumberOfTracks"},
                            );

            if ($main::AV{"$qf{zone},TransportState"} eq "PAUSED_PLAYBACK") {
                $template->param("ACTIVE_PAUSED"  => 1);
            } elsif ($main::AV{"$qf{zone},TransportState"} eq "STOPPED"){
                $template->param("ACTIVE_STOPPED" => 1);
            } else {
                $template->param("ACTIVE_PLAYING" => 1);
            }

            if ($main::AV{"$qf{zone},CurrentPlayMode"} eq "NORMAL") {
            } elsif ($main::AV{"$qf{zone},CurrentPlayMode"} eq "REPEAT_ALL") {
                $template->param("ACTIVE_REPEAT"   => 1);
            } elsif ($main::AV{"$qf{zone},CurrentPlayMode"} eq "SHUFFLE_NOREPEAT") {
                $template->param("ACTIVE_SHUFFLE"  => 1);
            } elsif ($main::AV{"$qf{zone},CurrentPlayMode"} eq "SHUFFLE") {
                $template->param("ACTIVE_SHUFFLE"  => 1);
                $template->param("ACTIVE_REPEAT"   => 1);
            }
        }

        $template->param("ACTIVE_VOLUME"  => $main::RENDER{"$qf{zone},Volume"}->{Master}->{val});

        if ($main::RENDER{"$qf{zone},Mute"}->{Master}->{val}) {
            $template->param("ACTIVE_MUTED"  => 1);
        }


    }

    my $all_arg = $r->uri->query;
    $all_arg =~ s/[&]*action=[^&]+//;
    $template->param("ALL_ARG" => $all_arg);

    if ($template->query(name => "QUEUE_LOOP") && $main::QUEUE{$qf{zone}} && @{$main::QUEUE{$qf{zone}}}[0]) {
        my @loop_data = ();
        foreach my $queue (@{@{$main::QUEUE{$qf{zone}}}[0]}) {
            my %qhash = %{$queue};
            my %row_data;
            if ($qhash{res}->{content} eq $main::AV{"$qf{zone},CurrentTrackMetaData"}->{item}->{res}->{content}) {
                if ($main::AV{"$qf{zone},TransportState"} eq "PLAYING") {
                    $row_data{QUEUE_IMG} = "playing.gif";
                } else {
                    $row_data{QUEUE_IMG} = "paused.gif";
                }
            } else {
                $row_data{QUEUE_IMG} = "blank.gif";
            }
            $row_data{QUEUE_NAME} = $qhash{"dc:title"};
            $row_data{QUEUE_ARG}  = "zone=$qf{zone}&queue=$qhash{id}";
            push(@loop_data, \%row_data);
        }
        $template->param("QUEUE_LOOP" => \@loop_data);
    }

    if ($template->query(name => "MUSIC_LOOP") && $main::QUEUE{$qf{zone}}) {
        my @loop_data = ();

        my $sortsub = sub {};

        if ($qf{msearch}) {
            foreach my $artist (sort {$a cmp $b} keys %main::MUSIC) {
                if ($artist =~ /$qf{msearch}/i) {
                    my %row_data;
                    $row_data{MUSIC_NAME} = $artist;
                    $row_data{MUSIC_ARG} = "zone=$qf{zone}&mpath=/" . encode_entities($artist);
                    push(@loop_data, \%row_data);
                }
                if ($#loop_data > $main::MAX_SEARCH) {
                    $template->param("MUSIC_ERROR" => "More then $main::MAX_SEARCH matching items.<BR>");
                    last;
                }
                foreach my $album (sort {$a cmp $b} keys %{$main::MUSIC{$artist}}) {
                    if ($album =~ /$qf{msearch}/i) {
                        my %row_data;
                        $row_data{MUSIC_NAME} = $album;
                        $row_data{MUSIC_ARG} = "zone=$qf{zone}&mpath=/" . encode_entities($artist) 
                                                                  . "/" . encode_entities($album);
                        push(@loop_data, \%row_data);
                    }
                    if ($#loop_data > $main::MAX_SEARCH) {
                        $template->param("MUSIC_ERROR" => "More then $main::MAX_SEARCH matching items.<BR>");
                        last;
                    }
                    foreach my $song (sort {$a cmp $b} keys %{$main::MUSIC{$artist}{$album}}) {
                        if ($song =~ /$qf{msearch}/i) {
                            my %row_data;
                            $row_data{MUSIC_NAME} = $song;
                            $row_data{MUSIC_ARG} = "zone=$qf{zone}&mpath=/" . encode_entities($artist) 
                                                                      . "/" . encode_entities($album)
                                                                      . "/" . encode_entities($song);
                            $row_data{MUSIC_ISSONG} = 1;
                            push(@loop_data, \%row_data);
                        }
                    }
                }
            }
            $template->param("MUSIC_LOOP" => \@loop_data);
            $template->param("MUSIC_SEARCH" => $qf{msearch});

        } else {
            my $mpath = $qf{mpath};
            my $mhash;
            my @parts = split("/", $mpath);

            if (int(@parts) == 2) {
                $mhash = \%{$main::MUSIC{$parts[1]}};
                $sortsub = sub {$a cmp $b};
            } elsif (int(@parts) == 3) {
                $mhash = \%{$main::MUSIC{$parts[1]}{$parts[2]}};
                $sortsub = sub {$mhash->{$a}{"upnp:originalTrackNumber"} <=> $mhash->{$b}{"upnp:originalTrackNumber"}};
            } else {
                $template->param("MUSIC_ROOT" => "1");
                $sortsub = sub {$a cmp $b};
                $mhash = \%main::MUSIC;
                $mpath = "";
            }

            foreach my $music (sort $sortsub keys %$mhash) {
                my %row_data;
                $row_data{MUSIC_NAME} = $music;
                $row_data{MUSIC_ARG} = "zone=$qf{zone}&mpath=$mpath/" . encode_entities($music);
                $row_data{MUSIC_ISSONG} = 1 if (int(@parts) == 3);
                push(@loop_data, \%row_data);
            }
            $template->param("MUSIC_LOOP" => \@loop_data);
        }
    }
    
    my $response = HTTP::Response->new(200, undef, [Connection => "close", "Content-Type" => "text/html; charset=ISO-8859-1"], $template->output);
    $c->send_response($response);
    $c->force_last_request;
    $c->close;
}

###############################################################################
sub reindex {

    # Pick the first coordinator's location to fetch everything from
    my @zones = keys (%main::ZONES);
    my $location = $main::ZONES{$main::ZONES{$zones[0]}->{Coordinator}}->{Location};

    my $contentDir = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:ContentDirectory:1");
    $main::REINDEX{contentDirProxy} = $contentDir->controlProxy; 
    $main::REINDEX{start} = 0;
    undef %main::REINDEX_DATA;
    Log (0, "Fetching DB");
    reindex_callback();
}

###############################################################################
sub reindex_callback {
    # We only do 100 at a time otherwise we lock out everything else
    my $result = $main::REINDEX{contentDirProxy}->Browse("A:TRACKS", 'BrowseDirectChildren', 
                                          'dc:title,res,dc:creator,upnp:artist,upnp:album', 
                                          $main::REINDEX{start}, 100, "");

    $main::REINDEX{start} += $result->getValue("NumberReturned");

    my $results = $result->getValue("Result");
    my $tree = XMLin($results, forcearray => ["item"]);

    foreach my $objectid (keys %{$tree->{item}}) {
        my $entry = $tree->{item}{$objectid};
        $entry->{"dc:creator"} = "Unknown" if (! $entry->{"dc:creator"});
        $entry->{"upnp:album"} = "Unknown" if (! $entry->{"upnp:album"});

        $main::REINDEX_DATA{$entry->{"dc:creator"}}{$entry->{"upnp:album"}}{$entry->{"dc:title"}} = $entry;
        $main::REINDEX_DATA{$entry->{"dc:creator"}}{$entry->{"upnp:album"}}{$entry->{"dc:title"}}{"objectid"} = {$objectid};
    }

    if ($main::REINDEX{start} >= $result->getValue("TotalMatches")) {
        %main::MUSIC = %main::REINDEX_DATA;
        undef %main::REINDEX_DATA;

        {
            local $Data::Dump::Purity = 1;
            Log (0, "Saving Music DB");
            open (DB, ">musicdb.pl");
            my $dumper = Data::Dumper->new( [\%main::MUSIC], [ qw( *main::MUSIC) ] );
            print DB $dumper->Dump();
            close DB;
            Log (0, "Finshed Saving Music DB");
        }
    } else {
        add_callback (time(), \&reindex_callback);
    }


}
###############################################################################
sub renew_subscriptions {
    foreach my $sub (keys %main::SUBSCRIPTIONS) {
        Log (2, "renew $sub");
        $main::SUBSCRIPTIONS{$sub}->renew;
    }
    add_callback(time()+$main::RENEW_SUB_TIME, \&renew_subscriptions);
}

###############################################################################
sub quit {
    foreach my $sub (keys %main::SUBSCRIPTIONS) {
        Log (1, "Unsubscribe $sub");
        $main::SUBSCRIPTIONS{$sub}->unsubscribe;
    }
    $main::daemon->close;
    Log (0, "Shutting Down");
    exit 0;
}

###############################################################################
sub get_service {
    my ($device, $name) = @_;

    my $service = $device->getService($name);
    return $service if ($service);

    for my $child ($device->children) {
        $service = $child->getService($name);
        return $service if ($service);
    }
    return undef;
}

###############################################################################
sub get_zone_id {
    my ($location) = @_;

    foreach my $zone (keys %main::ZONES) {
        return $zone if ($main::ZONES{$zone}->{Location} eq $location);
    }
    return undef;
}

###############################################################################
sub service_sub_cb {
    my ($service, %properties) = @_;

    my $zone = get_zone_id($service->{BASE});
    Log (1, "Event received for service=" . $service->{BASE}. " id=" . $service->serviceId . " zone=" . $zone);

    # Save off the zone names
    if ($service->serviceId =~ /serviceId:ZoneGroupTopology/) {
        my $tree = XMLin(decode_entities($properties{ZoneGroupState}), 
                forcearray => ["ZoneGroup"], 
                keyattr=>{"ZoneGroup" => "ID"});
        foreach my $key (keys %{$tree->{ZoneGroup}}) {
            my $zkey = $key;
            $zkey =~ s/:.*//g;
            $main::ZONES{$zkey} = $tree->{ZoneGroup}->{$key}->{ZoneGroupMember};
            $main::ZONES{$zkey}->{Coordinator} = $tree->{ZoneGroup}->{$key}->{Coordinator};
        }

        return;
    }

    # Save off the current status
    if ($service->serviceId =~ /serviceId:RenderingControl/) {
        my $tree = XMLin(decode_entities($properties{LastChange}), 
                forcearray => ["ZoneGroup"], 
                keyattr=>{"Volume"   => "channel",
                          "Mute"     => "channel",
                          "Loudness" => "channel"});
        foreach my $key ("Volume", "Treble", "Bass", "Mute", "Loudness") {
            if ($tree->{InstanceID}->{$key}) {
                $main::RENDER{"$zone,$key"} = $tree->{InstanceID}->{$key};
            }
        }
        return;
    }

    if ($service->serviceId =~ /serviceId:AVTransport/) {
        my $tree = XMLin(decode_entities($properties{LastChange}));

        foreach my $key ("CurrentTrackMetaData", "CurrentPlayMode", "NumberOfTracks", "CurrentTrack", "TransportState") {
            if ($tree->{InstanceID}->{$key}->{val}) {
                if ($tree->{InstanceID}->{$key}->{val} =~ /^</) {
                    $main::AV{"$zone,$key"} = XMLin($tree->{InstanceID}->{$key}->{val});
                } else {
                    $main::AV{"$zone,$key"} = $tree->{InstanceID}->{$key}->{val};
                }
            } else {
                undef $main::AV{"$zone,$key"};
            }
        }

        foreach my $data (@main::WAITINGAV) {
            send_http_response($data->[0], $data->[1], $data->[2]);
        }
        @main::WAITINGAV = ();

        return;
    }

    if ($service->serviceId =~ /serviceId:ContentDirectory/) {
        if (!defined $main::QUEUE{$zone} || $properties{ContainerUpdateIDs} =~ /Q:0/) {
            Log (1, "Refetching Q for $main::ZONES{$zone}->{ZoneName}");
            $main::QUEUE{$zone} = content_dir_browse($service->{BASE}, "Q:0");
        }
    }


    while (my ($key, $val) = each %properties) {
        if ($val =~ /&lt/) {
            my $d = decode_entities($val);
            my $tree = XMLin($d, forcearray => ["ZoneGroup"], keyattr=>{"ZoneGroup" => "ID"});
            Log(2, "Property ${key}'s value is " . Dumper($tree));
        } else {
            Log(2, "Property ${key}'s value is " . $val);
        }
    }
}

###############################################################################
sub content_dir_browse {
    my ($location, $objectid) = @_;

    my $start = 0;
    my @data = ();
    my $result;

    my $contentDir = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:ContentDirectory:1");
    my $contentDirProxy = $contentDir->controlProxy; 

    do {
        $result = $contentDirProxy->Browse($objectid, 'BrowseDirectChildren', 
                                           'dc:title,res,dc:creator,upnp:artist,upnp:album', 
                                           $start, 5000, "");

        return undef if (!$result->isSuccessful);

        $start += $result->getValue("NumberReturned");

        my $results = $result->getValue("Result");

        my $tree = XMLin($results, forcearray => ["item"], keyattr=>{"item" => "ID"});
        push(@data, $tree->{item}) if (defined $tree->{item});
        push(@data, $tree->{container}) if (defined $tree->{container});
    } while ($start < $result->getValue("TotalMatches"));

    return \@data;
}

###############################################################################
sub avtransport_remove_track {
    my ($location, $objectid) = @_;

    my $avTransport = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;

    my $result = $avTransportProxy->RemoveTrackFromQueue("0", $objectid);
    return;
}
###############################################################################
sub avtransport_play {
    my ($location) = @_;

    my $avTransport = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;

    my $result = $avTransportProxy->Play("0", "1");
    return $result;
}
###############################################################################
sub avtransport_seek {
    my ($location,$queue) = @_;

    my $avTransport = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;

    $queue =~ s,^.*/,,;

    my $result = $avTransportProxy->Seek("0", "TRACK_NR", $queue);
    return $result;
}
###############################################################################
sub render_mute {
    my ($location,$on) = @_;

    my $render = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:RenderingControl:1");
    my $renderProxy = $render->controlProxy;
    my $zone = get_zone_id($location);
    my $result = $renderProxy->SetMute("0", "Master", $on);
    $main::RENDER{"$zone,Mute"}->{Master}->{val} = $on;
    return $result;
}
###############################################################################
sub render_volume_change {
    my ($location,$change) = @_;

    my $zone = get_zone_id($location);
    my $vol = $main::RENDER{"$zone,Volume"}->{Master}->{val} + $change;
    render_volume($location,$vol);
}
###############################################################################
sub render_volume {
    my ($location,$vol) = @_;

    my $render = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:RenderingControl:1");
    my $renderProxy = $render->controlProxy;
    my $zone = get_zone_id($location);
    my $result = $renderProxy->SetVolume("0", "Master", $vol);
    $main::RENDER{"$zone,Volume"}->{Master}->{val} = $vol;
    return $result;
}
###############################################################################
sub avtransport_action {
    my ($location,$action) = @_;

    my $avTransport = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;

    my $result = $avTransportProxy->$action("0");
    return $result;
}
###############################################################################
sub avtransport_repeat {
    my ($location, $repeat) = @_;

    my $avTransport = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;
    my $zone = get_zone_id($location);
    my $str = $main::AV{"$zone,CurrentPlayMode"};

    if ($str eq "NORMAL") {
        $str = "REPEAT_ALL" if ($repeat);
    } elsif ($str eq "REPEAT_ALL") {
        $str = "NORMAL" if (!$repeat);
    } elsif ($str eq "SHUFFLE_NOREPEAT") {
        $str = "SHUFFLE" if ($repeat);
    } elsif ($str eq "SHUFFLE") {
        $str = "SHUFFLE_NOREPEAT" if (!$repeat);
    }
    my $result = $avTransportProxy->SetPlayMode("0", $str);
    $main::AV{"$zone,CurrentPlayMode"} = $str;
    return $result;
}
###############################################################################
sub avtransport_shuffle {
    my ($location, $shuffle) = @_;

    my $avTransport = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;
    my $zone = get_zone_id($location);
    my $str = $main::AV{"$zone,CurrentPlayMode"};

    if ($str eq "NORMAL") {
        $str = "SHUFFLE_NOREPEAT" if ($shuffle);
    } elsif ($str eq "REPEAT_ALL") {
        $str = "SHUFFLE" if ($shuffle);
    } elsif ($str eq "SHUFFLE_NOREPEAT") {
        $str = "NORMAL" if (!$shuffle);
    } elsif ($str eq "SHUFFLE") {
        $str = "REPEAT_ALL" if (!$shuffle);
    }

    my $result = $avTransportProxy->SetPlayMode("0", $str);
    $main::AV{"$zone,CurrentPlayMode"} = $str;
    return $result;
}
###############################################################################
sub avtransport_add {
    my ($location, $mpath) = @_;

    my @parts = split("/", $mpath);

    if (int(@parts) == 1) {
        foreach my $music (sort {$a cmp $b} keys %main::MUSIC) {
            avtransport_add($location, "/$music");
        }
        return;
    } elsif (int(@parts) == 2) {
        foreach my $music (sort {$a cmp $b} keys %{$main::MUSIC{$parts[1]}}) {
            avtransport_add($location, "$mpath/$music");
        }
        return;
    } elsif (int(@parts) == 3) {
        my $mhash = \%{$main::MUSIC{$parts[1]}{$parts[2]}};
        foreach my $music (sort {$mhash->{$a}{"upnp:originalTrackNumber"} <=> $mhash->{$b}{"upnp:originalTrackNumber"}} keys %{$mhash}) {
            avtransport_add($location, "$mpath/$music");
        }
        return;
    } 
    
    my $entry = $main::MUSIC{$parts[1]}{$parts[2]}{$parts[3]};
    my $avTransport = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;
    my $result = $avTransportProxy->AddURIToQueue(0, $entry->{res}->{content}, "");

    return;
}
###############################################################################
sub search_cb {
    my ($search, $device, $action) = @_;
    if ($action eq 'deviceAdded') {
        Log(1, "Device: " . $device->friendlyName . " added.");
        #print Dumper($device);

        $main::DEVICE{$device->{LOCATION}} = $device;
         
        my $zonegroup = get_service($device, "urn:schemas-upnp-org:service:ZoneGroupTopology:1");
        $main::SUBSCRIPTIONS{"$device->{LOCATION},zonegroup"} = $zonegroup->subscribe(\&service_sub_cb);

        my $contentDir = get_service($device, "urn:schemas-upnp-org:service:ContentDirectory:1");
        $main::SUBSCRIPTIONS{"$device->{LOCATION},contentDir"} = $contentDir->subscribe(\&service_sub_cb);
        $main::contentDirProxy = $contentDir->controlProxy; 

        my $avtransport = get_service($device, "urn:schemas-upnp-org:service:AVTransport:1");
        $main::SUBSCRIPTIONS{"$device->{LOCATION},avtransport"} = $avtransport->subscribe(\&service_sub_cb);

        my $rendercontrol = get_service($device, "urn:schemas-upnp-org:service:RenderingControl:1");
        $main::SUBSCRIPTIONS{"$device->{LOCATION},rendercontrol"} = $rendercontrol->subscribe(\&service_sub_cb);
    }
    elsif ($action eq 'deviceRemoved') {
        Log(1, "Device: " . $device->friendlyName . " removed.");
    } else {
        Log(1, "Device: " . $device->friendlyName . " unknown action: $action.");
    }
}
